/*
 *                            *
 *                           * *
 *                          *   *
 *                     ***************
 *                      * *       * *
 *                       *  MUMPS  *
 *                      * *       * *
 *                     ***************
 *                          *   *
 *                           * *
 *                            *
 *
 *   shmmgr.c
 *    shared memory manager
 *
 *  
 *   Author: John P. Willis <jpw@coherent-logic.com>
 *    Copyright (C) 1998 MUG Deutschland
 *    Copyright (C) 2020 Coherent Logic Development LLC
 *
 *
 *   This file is part of FreeM.
 *
 *   FreeM is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU Affero Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   FreeM is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU Affero Public License for more details.
 *
 *   You should have received a copy of the GNU Affero Public License
 *   along with FreeM.  If not, see <https://www.gnu.org/licenses/>.
 *
 **/

#include <stdlib.h>
#include <stdio.h>
#include "shmmgr.h"
#include "mpsdef.h"
#include <assert.h>

shm_config_t *shm_config;

short shm_init(const size_t seg_size)
{
    size_t alloc_map_size;

    register int i;
    
    shm_config = (shm_config_t *) malloc (sizeof (shm_config_t));
    NULLPTRCHK(shm_config,"shm_init");    
    
    /* figure out how many pages we can fit in the segment, accounting for header size */
    shm_config->pgct = (seg_size / PG_SIZE) - sizeof (shm_hdr_t);
    
    /* how big will the alloc map be? */
    alloc_map_size = shm_config->pgct * sizeof (shm_page_t);
    
    shm_config->segsiz = seg_size + alloc_map_size;
    shm_config->key = 0xBBDD20;
    shm_config->pgsiz = PG_SIZE;
    
    shm_config->seg_id = shmget (shm_config->key, shm_config->segsiz, 0644 | IPC_CREAT);
    if (shm_config->seg_id == -1) return SHMS_GET_ERR;    

    shm_config->dta = shmat (shm_config->seg_id, NULL, 0);
    if (shm_config->dta == (void *) -1) return SHMS_ATTACH_ERR;

    /* view the first sizeof (shm_hdr_t) bytes of the data area as an shm_hdr_t */
    shm_config->hdr = (shm_hdr_t *) shm_config->dta;
    
    if (shm_config->hdr->magic != shm_config->key) {

        /* zero out the segment */
        memset (shm_config->dta, 0, shm_config->segsiz);
        
        /* we are the process that created the segment: initialize everything */
        shm_config->hdr->magic = shm_config->key;
        shm_config->hdr->first_process = pid;

        /* store the address we got into the shm_hdr (borrowed from RSM) */
        shm_config->hdr->shmad = shm_config->dta;
        
        /* alloc_map comes after the header */
        shm_config->alloc_map = (shm_page_t *) (shm_config->dta + sizeof (shm_hdr_t));
        shm_config->buf = shm_config->alloc_map + ((shm_config->pgct * sizeof (shm_page_t)) + 1);
        
        for (i = 0; i < shm_config->pgct; i++) {
            shm_page_t *p = shm_get_alloc_map_entry (i);

            if (p != NULL) {
                p->is_first = FALSE;
                p->pg_state = PG_FREE;
                p->pid = 0;
                p->is_last = FALSE;
            }
        }
        
    }
    else {

        /* we are NOT the initial process. if addresses don't match, re-attach! */
        /* (again, borrowed from RSM) */
        if (shm_config->hdr->shmad != shm_config->dta) {

            /* grab the pointers we need */
            void *old_addr = shm_config->dta;
            void *new_addr = shm_config->hdr->shmad;

            /* detach and reattach */
            i = shmdt (old_addr);
            shm_config->dta = shmat (shm_config->seg_id, new_addr, 0);

            /* now, we need to fix up shm_config->alloc_map */
            shm_config->alloc_map = (shm_page_t *) (shm_config->dta + sizeof (shm_hdr_t));
            shm_config->buf = shm_config->alloc_map + ((shm_config->pgct * sizeof (shm_page_t)) + 1);
        }

    }    

    assert(shm_address_to_page_num(shm_page_num_to_address(20)) == 20);
    
}

shm_page_t *shm_get_alloc_map_entry(const int page_number)
{
    if ((page_number > (shm_config->pgct - 1)) || (page_number < 0)) return (shm_page_t *) NULL;
    
    return (shm_page_t *) (shm_config->alloc_map + (page_number * sizeof (shm_page_t)));
}

void *shm_page_num_to_address(const int page_num)
{
    return (void *) shm_config->buf + (shm_config->pgsiz * page_num);
}

int shm_address_to_page_num(const void *address)
{
    unsigned long val = (unsigned long) address - (unsigned long) shm_config->buf;
    unsigned long new_val = val / shm_config->pgsiz;
    
    return (int) new_val;
}

void *shm_alloc_pages(const int page_count)
{
    
    register int i;
    register int j;

    int candidate_page = 0;
    int free_pages_gotten = 0;
    
    shm_page_t *pg;
    
    for (i = 0; i < shm_config->pgct; i++) {

        pg = shm_get_alloc_map_entry (i);

        free_pages_gotten = 0;
        
        if (pg->pg_state == PG_FREE) {

            candidate_page = i;
            
            for (j = i; ((j < (i + page_count)) && (j < shm_config->pgct)); j++) {
                pg = shm_get_alloc_map_entry (j);

                if (pg->pg_state == PG_FREE) free_pages_gotten++;
                
            }

            if (free_pages_gotten == page_count) {

                for (j = candidate_page; j < (candidate_page + page_count); j++) {
                    pg = shm_get_alloc_map_entry (j);

                    pg->pg_state = PG_ALLOC;
                    pg->pid = pid;
                    
                    if (j == candidate_page) {
                        pg->is_first = TRUE;
                    }

                    if (j == candidate_page + (page_count - 1)) {
                        pg->is_last = TRUE;
                    }
                    
                }

                return (void *) shm_config->buf + (shm_config->pgsiz * candidate_page);
                
            }
            
        }

    }

    return (void *) NULL;
    
}


void *shm_alloc(const size_t bytes)
{
    int pages_needed = bytes / shm_config->pgsiz;
    float extra = bytes % shm_config->pgsiz;

    if (extra > 0) {
        pages_needed++;
    }

    return shm_alloc_pages (pages_needed);
}

void shm_free_page(const int page_number)
{
    register int i;
    shm_page_t *a = shm_get_alloc_map_entry (page_number);

    if (a->is_first == FALSE) {
        fprintf (stderr, "shm_free_page:  attempt to free page in the middle of allocation chain\r\n");
        return;
    }

    if (a->pg_state == PG_FREE) {
        fprintf (stderr, "shm_free_page:  double free attempted in page %d\r\n", page_number);
        exit (1);
    }

    for (i = page_number; i < shm_config->pgct; i++) {

        a = shm_get_alloc_map_entry (i);        
        
        if (a->is_last) {
            a->is_first = FALSE;
            a->pg_state = PG_FREE;
            a->pid = 0;
            a->is_last = FALSE;
            
            return;
        }
        else {
            a->is_first = FALSE;
            a->pg_state = PG_FREE;
            a->pid = 0;
            a->is_last = FALSE;
        }
        
    }
   
}

void shm_free(const void *addr)
{
    shm_free_page (shm_address_to_page_num (addr));
}

void *shmd_find(const short shmd_key)
{
    
    register int i;
    shm_dir_t *d;
    
    for (i = 0; i < SHM_DIRL; i++) {
        d = &(shm_config->hdr->shmdir[i]);

        if (d->shmd_key == shmd_key) return d->addr; 
    }

    return (void *) NULL;
}

void *shmd_insert(const short key, const void *addr)
{
    register int i;
    shm_dir_t *d;

    /* don't allow duplicates */
    for (i = 0; i < SHM_DIRL; i++) {
        if (d->shmd_key = key) return (void *) NULL;
    }
    
    for (i = 0; i < SHM_DIRL; i++) {
        d = &(shm_config->hdr->shmdir[i]);

        if (d->shmd_key == SHMDK_NONE) {
            d->shmd_key = key;
            d->addr = addr;

            return d;            
        }
        
    }

    return (void *) NULL;    
}

void shm_dump(void)
{

    printf ("SHARED MEMORY CONFIGURATION\r\n");
    printf ("  pgsiz                   %d\r\n", shm_config->pgsiz);
    printf ("  pgct                    %d\r\n", shm_config->pgct);
    printf ("  key                     %d\r\n", shm_config->key);
    printf ("  segid                   %d\r\n", shm_config->seg_id);
    printf ("  sizeof shm_page_t       %d\r\n", sizeof (shm_page_t));
    printf ("  segsiz                  %d\r\n", shm_config->segsiz);
    printf ("  shm address             %p\r\n", shm_config->dta);
    printf ("  alloc_map size          %d\r\n", sizeof (shm_page_t) * shm_config->pgct);
    printf ("  alloc_map address       %p\r\n", shm_config->alloc_map);
    printf ("  buf address             %p\r\n", shm_config->buf);
}

void shm_dump_pages(void)
{

    register int i;
    shm_page_t *p;

    printf ("%-10s%-10s%-10s%-10s%-10s\r\n", "PAGE", "PID", "BMHEAD", "BMTAIL", "STATE");
    
    for (i = 0; i < shm_config->pgct; i++) {

        p = shm_get_alloc_map_entry (i);

        printf ("%-10d%-10d%-10s%-10s%-10s\r\n",
                i,
                p->pid,
                (p->is_first == TRUE) ? "Y" : "N",
                (p->is_last == TRUE) ? "Y" : "N",
                (p->pg_state == PG_FREE) ? "PG_FREE" : "PG_ALLOC");
        
    }

}
